#!/usr/bin/env python3
# Autor: Jmeno Prijmeni

# Formát tohoto generátoru není nutné dodržovat nějak striktně, ale zavádí
# rozumné konvence. Pokud něčemu v kódu nerozumíš, neboj se někoho zeptat
# (autorem vzoru je Dan Skýpala)

from dataclasses import dataclass
from math import ceil, sqrt
from itertools import product
from random import *
import sys
from typing import Any, Callable, TypeVar

T = TypeVar('T')

#############################################################
# Dodatečné náhodné funkce
#############################################################


def randAlmost(to: int, min_part: float = 0.9) -> int:
    # Vrátí náhodný long long který je v nějakém smyslu "skoro tak velký" jako `to`,
    # konkrétně vrátí číslo v intervalu [floor(value*min_part), value].
    # Užitečné např. pro generování velikosti problému
    if to == 0:
        return 0
    assert to >= 1
    return randint(ceil(min_part * to), to)


def randBool(p: float = 0.5) -> bool:
    # Vrátí true s pravděpodobností p a false s pravděpodobností 1-p
    return not (random() > p)


def randDifferentNumbers(a: int, b: int, k: int) -> list[int]:
    # Vrátí k různých náhodných čísel v rozsahu [a, b]
    return sample(range(a, b + 1), k)

#############################################################
# Formát problému
#############################################################

# Pro konstanty, které se liší mezi obtížnostmi, můžeme použít takovýto formát
# MAX_N[diff] říká, jaké je maximální N pro danou obtížnost (1 nebo 2)
# TODO: upravit tuto ukazku pro svou ulohu
SUBTASKS = 6
T_MAX = [100, 3000, 10, 100, 500, 3000]
O_MAX = [100, 50000, 20, 300, 1000, 50000]
M_MIN = [-100_000, 1] + [-100_000] * 4
M_MAX = 100_000


class Problem:
    t: int
    o: int
    restrictions: list[tuple[int, int, int]]

    def __init__(self, shuffle_after: bool = True) -> None:
        self.restrictions = []
        self.shuffle_after = shuffle_after

    def confuse(self, subtask: int):
        if subtask == 0:
            return
        
        p = [i for i in range(self.t)]
        shuffle(p)
        for i, (a, b, m) in enumerate(self.restrictions):
            self.restrictions[i] = (p[a], p[b], m)

        if self.shuffle_after:
            shuffle(self.restrictions)

    def check(self, subtask: int):
        # Tato metoda není nutná, ale je užitečná jako základní kontrola:
        # zkontroluje, jestli problém odpovídá limitům stanoveným v zadání
        assert 1 <= self.t <= T_MAX[subtask]
        assert 1 <= self.o <= O_MAX[subtask]
        if subtask == 0:
            assert self.t == self.o

        for i, (a, b, m) in enumerate(self.restrictions):
            assert 0 <= a < self.t
            assert 0 <= b < self.t
            assert a != b
            assert M_MIN[subtask] <= m <= M_MAX
            if subtask == 0:
                assert b == self.restrictions[(i+1)%self.t][0]

    def write(self):
        print(self.t, self.o)
        for (a, b, m) in self.restrictions:
            print(a+1, b+1, m)


def all_edges(t):
    return [
        (a, b) if randBool() else (b, a)
        for a in range(t) for b in range(a+1, t)
    ]

def choose_and_remove(edges: list[T]) -> T:
    i = randint(0, len(edges)-1)
    edges[-1], edges[i] = edges[i], edges[-1]
    return edges.pop()

#############################################################
# Pomocné funkce generování
#############################################################


def fail(err):
    print(err, file=sys.stderr)
    sys.exit(1)


@dataclass
class InputType:
    name: str
    subtask: int
    repeat: int
    seeded: bool
    generate: Callable[..., Problem]
    args: dict[str, Any]
    def __str__(self):
        return f"{self.name} repeat={self.repeat} seeded={str(self.seeded).lower()}"

    def gen(self) -> Problem:
        return self.generate(self.subtask, **self.args)

inputs: dict[str, InputType] = {}


def generate_input(
    name: str, subtasks: int, repeat: int = 1, seeded: bool = True, args: dict[str, Any] = {}
):
    def dec(f: Callable[..., Problem]):
        for subtask in subtasks:
            full_name = f"{subtask+1:02}_{name}"
            if full_name in inputs:
                fail(f"Opakované jméno vstupů {name}")

            inputs[full_name] = InputType(full_name, subtask, repeat, seeded, f, args)

        return f  # Vracíme f, aby se funkce dala odekorovat víckrát

    return dec


#############################################################
# Funkce na generování vstupů
#############################################################


@generate_input(name="random", subtasks=range(1, SUBTASKS), repeat=1)
def gen_random(subtask: int):
    problem = Problem()
    problem.t = randAlmost(T_MAX[subtask])
    problem.o = randAlmost(O_MAX[subtask])

    edges = all_edges(problem.t)
    while len(problem.restrictions) < problem.o:
        a, b = choose_and_remove(edges)
        problem.restrictions.append((a, b, randint(M_MIN[subtask], M_MAX)))

    return problem


@generate_input(name="cycle", subtasks=[0], repeat=5)
@generate_input(name="cycle", subtasks=range(1, SUBTASKS), repeat=1)
def gen_cycle(subtask: int, max_bust: bool = False, max_min: bool = False, reverse: bool = False):
    problem = Problem(shuffle_after=False)
    problem.t = problem.o = randAlmost(T_MAX[subtask])
    for i in range(problem.t):
        problem.restrictions.append((
            i, (i+1) % problem.t,
            (M_MAX if max_min else M_MIN[subtask]) if max_bust else randint(M_MIN[subtask], M_MAX)
        ))
    
    if reverse:
        problem.restrictions.reverse()

    return problem

@generate_input(name="slow", subtasks=range(1, SUBTASKS), repeat=1)
def gen_path(subtask: int):
    problem = Problem(shuffle_after=True)
    problem.t = T_MAX[subtask]
    problem.o = O_MAX[subtask]
    groups = [(0, 1)]
    groups_edge_count = [0]

    edges = []
    while (last := groups[-1][1]) != problem.t:
        groups.append((last, min(
            last + int(sqrt(problem.t)),
            problem.t
        )))
        groups_edge_count.append(0)
        for a in range(*groups[-2]):
            for b in range(*groups[-1]):
                edges.append((len(groups)-1, a, b))

    if subtask != 1:
        for i in range(1, len(groups)):
            problem.restrictions.append((
                randint(groups[i][0], groups[i][1]-1),
                randint(groups[i-1][0], groups[i-1][1]-1),
                randint(M_MIN[subtask], int(0.9 * M_MIN[subtask]))
            ))

    while len(problem.restrictions) < problem.o:
        group, a, b = choose_and_remove(edges)
        groups_edge_count[group] += 1
        problem.restrictions.append((a, b, groups_edge_count[group] * (group+1)))

    return problem

for y, z in product([True, False], repeat=2):
    if not any([y, z]):
        continue
    generate_input(
        name=f"cycle_{'min' if y else 'max'}{'_rev' if z else ''}",
        subtasks=range(1, SUBTASKS),
        repeat=1,
        args={"max_bust": True, "max_min": y, "reverse": z}
    )(gen_cycle)


@generate_input(name="complete_graph", subtasks=range(1, SUBTASKS), repeat=2)
def gen_complete_graph(subtask: int):
    problem = Problem()
    problem.t = randAlmost(int(sqrt(O_MAX[subtask])))
    problem.o = problem.t * (problem.t - 1) // 2

    times = [randint(0, 10**9) for _ in range(problem.t)]

    for a in range(problem.t):
        for b in range(a+1, problem.t):
            if randBool():
                b, a = a, b
            if times[a] < times[b]:
                b, a = a, b

            problem.restrictions.append(
                (a, b, min(times[a] - times[b], M_MAX))
            )

    return problem


@generate_input(name="from_times", subtasks=range(1, SUBTASKS), repeat=3)
def gen_same(subtask: int):
    problem = Problem()
    problem.t = randAlmost(T_MAX[subtask])
    problem.o = randAlmost(O_MAX[subtask])
    times = [randint(0, 10**9) for _ in range(problem.t)]

    edges = all_edges(problem.t)
    while len(problem.restrictions) < problem.o:
        a, b = choose_and_remove(edges)

        if times[a] < times[b] + M_MIN[subtask]:
            edges.append((b, a))
            continue

        problem.restrictions.append((a, b, min(times[a] - times[b], M_MAX)))

    return problem


#############################################################
# Samotné generování vstupů
#############################################################

if __name__ == "__main__":
    if len(sys.argv) == 1:
        for inp in inputs.values():
            print(inp)
        quit()
    elif len(sys.argv) == 2:
        _, name = sys.argv
    elif len(sys.argv) == 3:
        _, name, hex_seed = sys.argv
        seed(hex_seed)
    else:
        fail(f"Nepovolený počet argumentů: {len(sys.argv)}")

    problem = inputs[name].gen()
    problem.confuse(inputs[name].subtask)
    problem.check(inputs[name].subtask)
    problem.write()
